# -*- coding: utf-8 -*-
"""!@package grass.pygrass.massages

@brief Temporal Framework GRASS C-library interface

Fast and exit-safe interface to GRASS C-library functions
using ctypes and multiprocessing


(C) 2013 by the GRASS Development Team
This program is free software under the GNU General Public
License (>=v2). Read the file COPYING that comes with GRASS
for details.

@author Soeren Gebbert
"""

import sys
from multiprocessing import Process, Lock, Pipe
import logging
from ctypes import *
import grass.lib.gis as libgis
import grass.lib.raster as libraster
import grass.lib.vector as libvector
import grass.lib.date as libdate
import grass.lib.raster3d as libraster3d
from datetime import datetime
from grass.script.core import *
from grass.script.raster import *
import time
import grass.temporal as tgis

###############################################################################

def test_exception(args):
    """!This function will return an exception to test the exception handling"
    """
    ret = [IOError("Oh no an exception"), None]
    return ret

def test_wrong_return_value(args):
    """!This function will return an exception to test the exception handling"
    """
    return None


def test_fatal_error(args):
    """!Call G_fatal_error()
    """
    libgis.G_fatal_error("A fatal error")
    # No return needed here

###############################################################################

def has_raster_timestamp(args):
    """!Check if the file based GRASS timestamp is present

       @param args name and mapset (args=["name", "mapset"])
       @return True if present, False otherwise

    """
    name = args[0]
    mapset = args[1]
    check = False
    if libgis.G_has_raster_timestamp(name, mapset) == 1:
        check = True

    ret = [None, check]

    return ret

###############################################################################

def read_raster_timestamp(args):
    """!Read the file based GRASS timestamp

       Please have a look at the documentation of G_read_raster_timestamp.

       The timestamps to be send are tuples of values:
           - relative time (start, end, unit), start and end are of type
             integer, unit is of type string.
           - absolute time (start, end), start and end are of type datetime

       The end time may be None in case of a time instance.

       @param args name and mapset (args=["name", "mapset"])

    """
    name = args[0]
    mapset = args[1]
    exception = None
    ts = libgis.TimeStamp()
    check = libgis.G_read_raster_timestamp(name, mapset, byref(ts))

    if check < 1:
        exception = IOError("Unable to read timestamp file "
                     "for raster map <%s>" % (name))

    dates = _convert_timestamp_from_grass(ts)
    ret = [exception, dates]
    return ret

###############################################################################

def write_raster_timestamp(args):
    """!Write the file based GRASS timestamp

       Please have a look at the documentation of G_write_raster_timestamp.

       @param args name and mapset (args=["name", "timestring"])
    """
    name = args[0]
    timestring = args[2]
    exception = None
    ts = libgis.TimeStamp()
    check = libgis.G_scan_timestamp(byref(ts), timestring)

    if check != 1:
        exception = IOError("Unable to convert the timestamp <%s>"%(timestring))

    check = libgis.G_write_raster_timestamp(name, byref(ts))

    if check == -1:
        exception = IOError("Unable to create timestamp file "
                     "for raster map <%s>" % (name))

    if check == -2:
        exception = IOError("Invalid datetime in timestamp for raster map <%s>" %
                     (name))

    ret = [exception, None]
    return ret

###############################################################################

def remove_raster_timestamp(args):
    """!Remove the file based GRASS timestamp from a raster map

       Please have a look at the documentation of G_remove_raster_timestamp.

       @param args name and mapset (args=["name", "mapset"])

    """
    name = args[0]
    mapset = args[1]
    exception = None
    check = libgis.G_remove_raster_timestamp(name, mapset)

    if check == -1:
        exception = IOError("Unable to remove timestamp for raster map <%s>" %
                     (name))

    ret = [exception, None]
    return ret


###############################################################################

def raster_map_exists(args):
    """!Check if the raster map exists in the spatial database

       @param args name and mapset (args=["name", "mapset"])
       @return True if present, False otherwise

    """
    name = args[0]
    mapset = args[1]
    check = False
    mapset = libgis.G_find_raster(name, mapset)

    if mapset:
        check = True

    ret = [None, check]
    return ret

###############################################################################

def read_raster_info(args):
    """!Read the raster map info from the file system and store the content
       into a dictionary

       This method uses the ctypes interface to the gis and raster libraries
       to read the map metadata information

       @param args name and mapset (args=["name", "mapset"])
       @return The key value pairs of the map specific metadata,
               or None in case the map does not exists
    """

    name = args[0]
    mapset = args[1]
    exception = None

    kvp = {}

    if not libgis.G_find_raster(name, mapset):
        return

    # Read the region information
    region = libgis.Cell_head()
    libraster.Rast_get_cellhd(name, mapset, byref(region))

    kvp["north"] = region.north
    kvp["south"] = region.south
    kvp["east"] = region.east
    kvp["west"] = region.west
    kvp["nsres"] = region.ns_res
    kvp["ewres"] = region.ew_res
    kvp["rows"] = region.cols
    kvp["cols"] = region.rows

    maptype = libraster.Rast_map_type(name, mapset)

    if maptype == libraster.DCELL_TYPE:
        kvp["datatype"] = "DCELL"
    elif maptype == libraster.FCELL_TYPE:
        kvp["datatype"] = "FCELL"
    elif maptype == libraster.CELL_TYPE:
        kvp["datatype"] = "CELL"

    # Read range
    if libraster.Rast_map_is_fp(name, mapset):
        range = libraster.FPRange()
        libraster.Rast_init_fp_range(byref(range))
        ret = libraster.Rast_read_fp_range(name, mapset, byref(range))
        if ret < 0:
            exception = IOError("Unable to read range file")
        if ret == 2:
            kvp["min"] = None
            kvp["max"] = None
        else:
            min = libgis.DCELL()
            max = libgis.DCELL()
            libraster.Rast_get_fp_range_min_max(
                byref(range), byref(min), byref(max))
            kvp["min"] = min.value
            kvp["max"] = max.value
    else:
        range = libraster.Range()
        libraster.Rast_init_range(byref(range))
        ret = libraster.Rast_read_range(name, mapset, byref(range))
        if ret < 0:
            exception = IOError("Unable to read range file")
        if ret == 2:
            kvp["min"] = None
            kvp["max"] = None
        else:
            min = libgis.CELL()
            max = libgis.CELL()
            libraster.Rast_get_range_min_max(
                byref(range), byref(min), byref(max))
            kvp["min"] = min.value
            kvp["max"] = max.value

    ret = [exception, kvp]
    return ret


###############################################################################

def _convert_timestamp_from_grass(ts):
    """!Convert a GRASS file based timestamp into the temporal framework
       format datetime or integer.

       A tuple of two datetime objects (start, end) is returned in case of absolute time.
       In case of relative time a tuple with start time, end time and the
       relative unit (start, end, unit) will be returned.

       Note:
       The end time will be set to None in case of a time instance.

       @param ts grass.lib.gis.TimeStamp object created by G_read_*_timestamp
    """

    dt1 = libgis.DateTime()
    dt2 = libgis.DateTime()
    count = c_int()

    libgis.G_get_timestamps(byref(ts),
                            byref(dt1),
                            byref(dt2),
                            byref(count))


    if dt1.mode == libdate.DATETIME_ABSOLUTE:
        pdt1 = None
        pdt2 = None
        if count.value >= 1:
            pdt1 = datetime(int(dt1.year), int(dt1.month), int(dt1.day),
                            int(dt1.hour), int(dt1.minute),
                            int(dt1.second))
        if count.value == 2:
            pdt2 = datetime(int(dt2.year), int(dt2.month), int(dt2.day),
                            int(dt2.hour), int(dt2.minute),
                            int(dt2.second))

        # ATTENTION: We ignore the time zone
        # TODO: Write time zone support
        ret = (pdt1, pdt2)
        return ret
    else:
        unit = None
        start = None
        end = None
        if count >= 1:
            if dt1.year > 0:
                unit = "years"
                start = dt1.year
            elif dt1.month > 0:
                unit = "months"
                start = dt1.month
            elif dt1.day > 0:
                unit = "days"
                start = dt1.day
            elif dt1.hour > 0:
                unit = "hours"
                start = dt1.hour
            elif dt1.minute > 0:
                unit = "minutess"
                start = dt1.minutes
            elif dt1.seconds > 0:
                unit = "seconds"
                start = dt1.seconds
        if count == 2:
            if dt2.year > 0:
                end = dt2.year
            elif dt2.month > 0:
                end = dt2.month
            elif dt2.day > 0:
                end = dt2.day
            elif dt2.hour > 0:
                end = dt2.hour
            elif dt2.minute > 0:
                end = dt2.minutes
            elif dt2.seconds > 0:
                end = dt2.seconds
        ret = (start, end, unit)
        return ret

###############################################################################

def rpc_server_function(lock, conn):
    """!The GRASS C-libraries server function designed to be a target for
       multiprocessing.Process

       @param lock A multiprocessing.Lock
       @param conn A multiprocessing.Pipe
    """

    libgis.G_gisinit("c_library_server")

    while True:
        # Avoid busy waiting
        conn.poll(None)
        data = conn.recv()
        function = data[0]
        args = data[1]
        lock.acquire()
        ret = function(args)
        conn.send(ret)
        lock.release()

###############################################################################

class RPCServer(object):
    """!Fast and exit-safe interface to call Python functions
       that make use GRASS C-libraries functions using ctypes

       This class implements a fast and exit-safe interface to call Python
       functions that make use GRASS C-libraries functions using ctypes.

       The C-libraries functions are called via ctypes in a subprocess
       using a pipe (multiprocessing.Pipe) to transfer the functions
       and arguments.
       Hence, the process that uses the RPCServer will not be
       exited, if a G_fatal_error() was invoked in the subprocess.
       In this case the RPCServer object will simply start a
       new subprocess and restarts the pipeline.

       An IOException will be raised in case the server terminated.


       Usage:

       @code
       >>> import grass.script as grass
       >>> import grass.temporal as tgis
       >>> grass.use_temp_region()
       >>> grass.run_command("g.region", n=80.0, s=0.0, e=120.0, w=0.0,
       ... t=1.0, b=0.0, res=10.0, res3=10.0)
       0
       >>> tgis.init()
       >>> grass.run_command("r.mapcalc", expression="test = 1", overwrite=True, quiet=True)
       0
       >>> grass.run_command("r.timestamp", map="test", date='12 Mar 1995', overwrite=True, quiet=True)
       0

       # Raster map
       >>> ciface = RPCServer()
       >>> check = ciface.call(function=raster_map_exists, args=("test", tgis.get_current_mapset()))
       >>> print check
       True

       >>> ciface.call(function=test_fatal_error, args=(None))
       Traceback (most recent call last):
         File "c_library_interface.py", line 448, in cally
           raise IOError("Server subprocess exited with signal %i"%(self.server.exitcode))
       IOError: Server subprocess exited with signal 1

       >>> ciface.call(function=test_exception, args=(None))
       Traceback (most recent call last):
         File "c_library_interface.py", line 490, in call
           raise ret[0]
       IOError: Oh no an exception

       >>> ciface.call(function=read_raster_info, args=("test", tgis.get_current_mapset()))
       {'rows': 12, 'north': 80.0, 'min': 1, 'datatype': 'CELL', 'max': 1, 'ewres': 10.0, 'cols': 8, 'west': 0.0, 'east': 120.0, 'nsres': 10.0, 'south': 0.0}
       >>> check = ciface.call(function=has_raster_timestamp, args=("test", tgis.get_current_mapset()))
       >>> print check
       True
       >>> if check:
       ...     res = ciface.call(function=read_raster_timestamp, args=("test", tgis.get_current_mapset()))
       ...     if res[0]:
       ...         print str(res[0])
       ...         ciface.call(function=remove_raster_timestamp, args=("test", tgis.get_current_mapset()))
       1995-03-12 00:00:00

       >>> ciface.call(function=has_raster_timestamp, args=("test", tgis.get_current_mapset()))
       False
       
       >>> ciface.call(function=test_fatal_error, args=(None))
       Traceback (most recent call last):
         File "c_library_interface.py", line 448, in cally
           raise IOError("Server subprocess exited with signal %i"%(self.server.exitcode))
       IOError: Server subprocess exited with signal 1

       >>> ciface.call(function=write_raster_timestamp, args=("test", tgis.get_current_mapset(), "13 Jan 1999"))

       >>> ciface.call(function=test_exception, args=(None))
       Traceback (most recent call last):
         File "c_library_interface.py", line 490, in call
           raise ret[0]
       IOError: Oh no an exception

       >>> ciface.call(function=has_raster_timestamp, args=("test", tgis.get_current_mapset()))
       True

       >>> ciface.call(function=test_wrong_return_value, args=(None))
       Traceback (most recent call last):
         File "c_library_interface.py", line 501, in call
           raise IOError("Return value of called functions has the wrong layout")
       IOError: Return value of called function <test_wrong_return_value> has the wrong layout

       >>> grass.del_temp_region()

       @endcode
    """
    def __init__(self):
        self.client_conn = None
        self.server_conn = None
        self.server = None
        self._start_server()

    def __del__(self):
        self.stop()

    def _start_server(self):
        self.lock = Lock()
        self.client_conn, self.server_conn = Pipe()
        self.server = Process(target=rpc_server_function, args=(self.lock,
                                                          self.server_conn))
        self.server.daemon = True
        self.server.start()

    def _check_restart_server(self):
        """!Restart the server if it was terminated
        """
        if self.server.is_alive() is True:
            return
        self.client_conn.close()
        self.server_conn.close()
        self._start_server()
        logging.warning("Needed to restart the rpc server")

    def call(self, function, args):
        """!Call a Python function in the server subprocess

           @param function The function to execute
           @param args The arguments for the function as tuple
           @return The return value of the called function
       """
        self._check_restart_server()
        self.client_conn.send([function, args])

        # Wait up to two seconds to receive the data
        self.client_conn.poll(2)
        if self.server.is_alive() is False:
            raise IOError("Server subprocess exited with signal %i"%(self.server.exitcode))

        # Receive the data from the subprocess
        ret = self.client_conn.recv()

        if not ret or len(ret) < 2:
            raise IOError("Return value of called function <%s> has the wrong layout"%(str(function.__name__)))

        # Raise the exception if present in the return list
        if ret[0] is not None:
            raise ret[0]

        # Return the return vales of the called function
        return ret[1]

    def stop(self):
        """!Stop the messenger server and close the pipe
        """
        if self.server is not None and self.server.is_alive():
            self.server.terminate()
        if self.client_conn is not None:
            self.client_conn.close()
     
###############################################################################

def raster_exists_benchmark():    
    print "##############################################################"
    print "Raster map exists benchmark"

    num = 1000

    # Direct function calls
    start = time.time()
    for i in xrange(num):
        args = ("test", tgis.get_current_mapset())
        check = raster_map_exists(args)
    end = time.time()
    print "Time to call %i functions directly: %fs"%(num, end - start)

    # RPC server
    rpc = RPCServer()
    start = time.time()
    for i in xrange(num):
        args = ("test", tgis.get_current_mapset())
        kvp = rpc.call(function=raster_map_exists, args=args)
    end = time.time()
    print "Time to call %i functions via RPC: %fs"%(num, end - start)
    rpc.stop()

    # RPC server
    rpc = RPCServer()
    start = time.time()
    for i in xrange(num):
        kvp = find_file("test")
    end = time.time()
    print "Time to perform %i g.findfile module runs: %fs"%(num, end - start)
    rpc.stop()
    print "##############################################################"

def raster_info_benchmark():    
    print "##############################################################"
    print "Raster map info benchmark"

    # Benchmark
    num = 10000

    # Direct function calls
    start = time.time()
    for i in xrange(num):
        args = ("test", tgis.get_current_mapset())
        kvp = read_raster_info(args)
    end = time.time()
    print "Time to call %i functions directly: %fs"%(num, end - start)

    # RPC server
    rpc = RPCServer()
    start = time.time()
    for i in xrange(num):
        args = ("test", tgis.get_current_mapset())
        kvp = rpc.call(function=read_raster_info, args=args)
    end = time.time()
    print "Time to call %i functions via RPC: %fs"%(num, end - start)
    rpc.stop()

    # RPC server
    rpc = RPCServer()
    start = time.time()
    for i in xrange(num):
        kvp = raster_info("test")
    end = time.time()
    print "Time to perform %i r.info module runs: %fs"%(num, end - start)
    rpc.stop()
    print "##############################################################"

###############################################################################

if __name__ == "__main__":
    import doctest
    print "##############################################################"
    print "DOCTESTS:"
    doctest.testmod()
    print "##############################################################"

    # Start the benchmarks
    raster_exists_benchmark()
    raster_info_benchmark()    


