# -*- coding: utf-8 -*-
"""!@package grass.pygrass.rpc

@brief PyGRASS RPC server

(C) 2013 by the GRASS Development Team
This program is free software under the GNU General Public
License (>=v2). Read the file COPYING that comes with GRASS
for details.

@author Soeren Gebbert
"""

from multiprocessing import Process, Lock, Pipe
#import unittest
import logging

def rpc_server_function(lock, conn, obj_type, kargs):
    """!The GRASS Python object server function designed to be a target for
       multiprocessing.Process

       @param lock A multiprocessing.Lock
       @param conn A multiprocessing.Pipe
       @param obj_type The type of the object to be created,
              for example: grass.pygrass.raster.RasterRow
       @param kargs The keyword arguments for the object constructor as
              dictionary, example: {"name":"test"}
    """

    # Create the object that should be accessed remotely
    obj = obj_type(**kargs)

    while True:
        # Avoid busy waiting
        conn.poll(None)
        # Receive data
        data = conn.recv()
        is_method = data[0]
        lock.acquire()

        # We catch any exception and return it to the client
        exception = None

        # Call the method with arguments or a property of the object
        # Catch the exception and store an instance of it
        try:
            if is_method is True:
                method = data[1]
                method_kargs = data[2]
                ret = getattr(obj, method)(**method_kargs)
            else:
                prop = data[1]
                ret = getattr(obj, prop)
        except Exception as inst:
            exception = inst

        # Send the data
        conn.send([exception, ret])

        lock.release()

###############################################################################

class RPCServer(object):
    """!Fast and exit-safe interface to access Python objects remotely

       Usage:

       @code


       @endcode
    """
    def __init__(self, obj_type, kargs):
        """!Start the RPC server and create the persistant object in the
           remote process

        @param obj_type The type of the object to be created,
              for example: grass.pygrass.raster.RasterRow
        @param kargs The keyword arguments for the object constructor as
              dictionary, example: {"name":"test"}
        """
        self.client_conn = None
        self.server_conn = None
        self.server = None
        self.obj_type = obj_type
        self.obj_kargs = kargs
        self._start_server()

    def __del__(self):
        self.stop()

    def _start_server(self):
        self.lock = Lock()
        self.client_conn, self.server_conn = Pipe()
        self.server = Process(target=rpc_server_function, args=(self.lock,
                                                          self.server_conn,
                                                          self.obj_type,
                                                          self.obj_kargs))
        self.server.daemon = True
        self.server.start()

    def _check_restart_server(self):
        """!Restart the server if it was terminated
        """
        if self.server.is_alive() is True:
            return
        self.client_conn.close()
        self.server_conn.close()
        self._start_server()
        logging.warning("Needed to restart the rpc server")

    def call(self, method, kargs):
        """!Call a method of the remote object and return its return value

           @param method The method name of the remote object to be called as astring
           @param kargs The keyword arguments for the method as dictionary
           @return The return value of the called function
       """
        self._check_restart_server()
        self.client_conn.send([True, method, kargs])

        # Wait up to two seconds to receive the data
        self.client_conn.poll(2)
        if self.server.is_alive() is False:
            raise IOError("Server subprocess exited with signal %i"%(self.server.exitcode))

        # Receive the data from the subprocess
        ret = self.client_conn.recv()

        # Raise the exception if present in the return list
        if ret[0] is not None:
            raise ret[0]

        # Return the return vales of the called function
        return ret[1]

    def get_property(self, prop):
        """!Get the value of a property of the remote object

           @param prop The property name as string
           @return The return value of the property
       """
        self._check_restart_server()
        self.client_conn.send([False, prop])

        # Wait up to two seconds to receive the data
        self.client_conn.poll(2)
        if self.server.is_alive() is False:
            raise IOError("Server subprocess exited with signal %i"%(self.server.exitcode))

        # Receive the data from the subprocess
        ret = self.client_conn.recv()

        # Raise the exception if present in the return list
        if ret[0] is not None:
            raise ret[0]

        # Return the return vales of the called function
        return ret[1]

    def stop(self):
        """!Stop the messenger server and close the pipe
        """
        if self.server is not None and self.server.is_alive():
            self.server.terminate()
        if self.client_conn is not None:
            self.client_conn.close()


if __name__ == '__main__':
    import grass.script as grassscript
    import grass.pygrass.raster as pygrass

    grassscript.use_temp_region()
    grassscript.run_command("g.region", n=80.0, s=0.0, e=120.0, w=0.0,
                      t=1.0, b=0.0, res=10.0, res3=10.0)
    grassscript.run_command("r.mapcalc", expression="test = row()",
                            overwrite=True, quiet=True)

    # Start the server and create a RasterRow remote object
    # that is initialized with the raster map name
    kargs = {"name":"test"}
    server = RPCServer(pygrass.RasterRow, kargs)

    # Get some raster map info
    info = server.get_property("info")
    print "Rows:", info.rows
    print "Cols:", info.cols
    print "N:", info.north
    print "S:", info.south
    print "E:", info.east
    print "W:", info.west

    # Open the raster map for row access
    kargs = {}
    server.call("open", kargs)

    # Read the first row
    for row in xrange(info.rows):
        kargs = {"row":row}
        ret = server.call("get_row", kargs)
        print "Row:", row, ret

    # Close the raster map
    kargs = {}
    server.call("close", kargs)

    grassscript.del_temp_region()